;; Generated package description from calibre.el  -*- no-byte-compile: t -*-
(define-package "calibre" "1.3.4.0.20231212.110229" "Interact with Calibre libraries from Emacs" '((emacs "27.1") (compat "29.1.4.1")) :commit "2c4221ecfaf511cbca0205ccb3531d70246a71dc" :authors '(("Kjartan Oli Agustsson" . "kjartanoli@disroot.org")) :maintainer '("Kjartan Oli Agustsson" . "kjartanoli@disroot.org") :url "https://git.disroot.org/kjartanoli/calibre.el")
